import sys, os, time
import numpy as np
from .mytools.MinTree import MinTree

from .._model import DMmodel
from ...util.basicutil import param_default
from ...backend import STensor

def del_block(M, rowSet ,colSet):
    M = M.tolil()

    (rs, cs) = M.nonzero()
    for i in range(len(rs)):
        if rs[i] in rowSet or cs[i] in colSet:
            M[rs[i], cs[i]] = 0
    return M.tolil()


class FlowScope( DMmodel ):
    '''Anomaly detection base on contrastively dense subgraphs, considering
    topological, temporal, and categorical (e.g. rating scores) signals, or
    any supported combinations.

    Parameters
    ----------
    graph: Graph
        Graph instance contains adjency matrix, and possible multiple signals.
    '''
    def __init__(self, graphList: list, **params):
        self.graphnum = len(graphList)
        self.graphlist = graphList
        # self.alpha = param_default(params, 'alpha', 0.8)
        # self.alg = param_default(params, 'alg', 'fastgreedy')

    def __str__(self):
        return str(vars(self))

    
    def run(self, k:int=3, level:int=0, alpha:float=4):
        print("you are running with ", self.graphnum+1," partite graph")
        self.level = level
        self.alpha = alpha
        self.initData()
        self.nres = []


        for i in range(k):
            if self.level == 0:
                (finalsets, score) = self.fastGreedyDecreasing()
            else:
                return print("No such level know")

            self.nres.append([finalsets, score])

            for i in range(len(self.mcurlist)):
                self.mcurlist[i] = del_block(self.mcurlist[i], finalsets[i], finalsets[i+1]) 

        return self.nres


    def initData(self):
        self.mcurlist = []
        self.mtranslist = []
        for i in range(len(self.graphlist)):
            self.mcurlist.append(self.graphlist[i].graph_tensor._data.copy().tocsr().tolil().astype(np.float64))
            self.mtranslist.append(self.graphlist[i].graph_tensor._data.copy().tocsr().tolil().transpose()) 


    def initGreedy(self):
        self.sets = []
        self.dtrees = []
        self.deltaslist = []
        self.curScorelist = []
        self.curAveScorelist = []


        self.sets.append(set(range(self.mcurlist[0].shape[0])))
        for i in range(len(self.mcurlist)-1):
            self.sets.append(set(range(self.mcurlist[i].shape[1])))
        self.sets.append(set(range(self.mcurlist[1].shape[1])))
        
        
        rowDeltas = np.squeeze(self.mcurlist[0].sum(axis=1).A)  # sum of A
        self.dtrees.append(MinTree(rowDeltas))
        for i in range(len(self.mcurlist)-1):
            midDeltas1 = np.squeeze(self.mcurlist[i].sum(axis=0).A)
            midDeltas2 = np.squeeze(self.mcurlist[i+1].sum(axis=1).A)
            self.deltaslist.append(midDeltas1)
            self.deltaslist.append(midDeltas2)
            
            mid_min = []
            mid_max = []
            for (m1, m2) in zip(midDeltas1, midDeltas2):
                temp = min(m1, m2)  # fi
                temp2 = max(m1, m2)  # qi
                mid_min.append(temp)
                mid_max.append(temp2)

            mid_min = np.array(mid_min)
            mid_max = np.array(mid_max)
            mid_priority = (1 + self.alpha) * mid_min - self.alpha * mid_max
            self.dtrees.append(MinTree(mid_priority))

            curScore1 = sum(mid_min)
            curScore2 = sum(abs(midDeltas1 - midDeltas2))
            self.curScorelist.append(curScore1)
            self.curScorelist.append(curScore2)
            s = self.mcurlist[i].shape[0] + self.mcurlist[i].shape[1] + self.mcurlist[i+1].shape[1]
            curAveScore1 = curScore1 / s
            curAveScore2 = curScore2 / s
            self.curAveScorelist.append(curAveScore1)
            self.curAveScorelist.append(curAveScore2)

        colDeltas = np.squeeze(self.mcurlist[-1].sum(axis=0).A)  # sum of C
        self.dtrees.append(MinTree(colDeltas))
        
        

    def updataConnNode(self, mold, index, mindelta):
        if mold == 0:
            # update  the  weight of connected nodes
            for j in self.mcurlist[0].rows[mindelta]:
                
                if self.deltaslist[0][j] == -1:
                    continue
                new_md1 = self.deltaslist[0][j] - self.mcurlist[0][mindelta, j]
                tempmin = min(self.deltaslist[0][j], self.deltaslist[1][j])
                if (new_md1) < tempmin:  # if in-degree know is the smallist one, update the curScore1
                    self.curScorelist[0] -= (tempmin - new_md1)

                self.curScorelist[1] += abs(new_md1 - self.deltaslist[1][j]) - abs(self.deltaslist[0][j] - self.deltaslist[1][j])
                self.deltaslist[0][j] = new_md1

                # update mid_tree
                mid_delta_value = abs(self.deltaslist[0][j] - self.deltaslist[1][j])
                new_mid_w = tempmin - self.alpha * mid_delta_value
                self.dtrees[1].setVal(j, new_mid_w)

            self.sets[0] -= {mindelta}  # update rowSet
            self.dtrees[0].changeVal(mindelta, float('inf'))
            self.deleted.append((index, mindelta)) 
            self.numDeleted += 1

        elif mold == 2:
            # update mD1, mD2, and mid_tree
            for i in self.mtranslist[-1].rows[mindelta]:
                if self.deltaslist[-1][i] == -1:
                    continue

                new_md2 = self.deltaslist[-1][i] - self.mtranslist[-1][mindelta, i]
                tempmin = min(self.deltaslist[-1][i], self.deltaslist[-2][i])
                if (new_md2) < tempmin:
                    self.curScorelist[-2] -= (tempmin - new_md2)

                self.curScorelist[-1] += abs(new_md2 - self.deltaslist[-2][i]) - abs(self.deltaslist[-1][i] - self.deltaslist[-2][i])
                self.deltaslist[-1][i] = new_md2

                mid_delta_value = abs(self.deltaslist[-1][i] - self.deltaslist[-2][i])
                new_mid_w = tempmin - self.alpha * mid_delta_value
                self.dtrees[-2].setVal(i, new_mid_w)

            self.sets[-1] -= {mindelta}
            self.dtrees[-1].changeVal(mindelta, float('inf'))
            self.deleted.append((index, mindelta))
            self.numDeleted += 1

        elif mold == 1:

            self.curScorelist[2 * index - 2] -= min(self.deltaslist[2 * index - 2][mindelta],
                                                    self.deltaslist[2 * index - 1][mindelta])
            self.curScorelist[2 * index - 1] -= abs(
                self.deltaslist[2 * index - 2][mindelta] - self.deltaslist[2 * index - 1][mindelta])

            self.deltaslist[2 * index - 2][mindelta] = -1
            self.deltaslist[2 * index - 1][mindelta] = -1

            for j in self.mcurlist[index].rows[mindelta]:
                if index < (len(self.dtrees) - 2):
                    if self.deltaslist[2 * index][j] == -1:
                        continue
                    new_md1 = self.deltaslist[2 * index][j] - self.mcurlist[index][mindelta, j]
                    tempmin = min(self.deltaslist[2 * index][j], self.deltaslist[2 * index + 1][j])
                    if (new_md1) < tempmin:  # if in-degree know is the smallist one, update the curScore1
                        self.curScorelist[2 * index] -= (tempmin - new_md1)

                    self.curScorelist[2 * index + 1] += abs(new_md1 - self.deltaslist[2 * index + 1][j]) - abs(
                        self.deltaslist[2 * index][j] - self.deltaslist[2 * index + 1][j])
                    self.deltaslist[2 * index][j] = new_md1


                    mid_delta_value = abs(self.deltaslist[2 * index][j] - self.deltaslist[2 * index + 1][j])
                    new_mid_w = tempmin - self.alpha * mid_delta_value
                    self.dtrees[index + 1].setVal(j, new_mid_w)
                else:
                    self.dtrees[index + 1].changeVal(j, -self.mcurlist[index][mindelta, j])

            for j in self.mtranslist[index - 1].rows[mindelta]:
                if index > 1:
                    if self.deltaslist[2 * index - 3][j] == -1:
                        continue

                    new_md2 = self.deltaslist[2 * index - 3][j] - self.mtranslist[index - 1][mindelta, j]
                    tempmin = min(self.deltaslist[2 * index - 3][j], self.deltaslist[2 * index - 4][j])
                    if (new_md2) < tempmin:
                        self.curScorelist[2 * index - 4] -= (tempmin - new_md2)

                    self.curScorelist[2 * index - 3] += abs(new_md2 - self.deltaslist[2 * index - 4][j]) - abs(
                        self.deltaslist[2 * index - 3][j] - self.deltaslist[2 * index - 4][j])
                    self.deltaslist[2 * index - 3][j] = new_md2

                    mid_delta_value = abs(self.deltaslist[2 * index - 3][j] - self.deltaslist[2 * index - 4][j])
                    new_mid_w = tempmin - self.alpha * mid_delta_value
                    self.dtrees[index - 1].setVal(j, new_mid_w)
                else:
                    self.dtrees[index - 1].changeVal(j, -self.mtranslist[index - 1][mindelta, j])

            self.sets[index] -= {mindelta}
            self.dtrees[index].changeVal(mindelta, float('inf'))
            self.deleted.append((index, mindelta))
            self.numDeleted += 1
            
    
    # find the smallest one in all set
    def findmin(self):
        resmin = []
        for i in range(len(self.dtrees)):
            (next, delt) = self.dtrees[i].getMin()
            resmin.append((next, delt))

        min1 = resmin[0][1] * (1 + self.alpha) 
        min2 = resmin[-1][1] * (1 + self.alpha)
        min_weight = min1
        minidx = resmin[0][0]
        idx = 0
        mold = 0
        
        if min2 < min_weight:
            min_weight = min2
            idx = len(resmin) - 1
            mold = 2
            minidx = resmin[-1][0]
        for i in range(len(resmin) - 2):
            if resmin[i + 1][1] < min_weight:
                min_weight = resmin[i + 1][1]
                minidx = resmin[i + 1][0]
                idx = i + 1
                mold = 1
        
        return mold, idx, minidx
        
        
    def checkset(self):
        res = True
        for i in range(len(self.sets)):
            res = self.sets[i] and res
        return res

    
    # use to print the result of algorithm
    def printres(self):
        print('best delete number : ', self.bestNumDeleted)
        res1 = 'nodes number remaining:  '
        for i in range(len(self.sets)):
            res1 += str(len(self.sets[i])) + ' '
        print(res1)
        res2 = 'matrix mass remaining:  '
        for i in range(len(self.curScorelist)):
            res2 += str((self.curScorelist[i])) + ' '
        print(res2)

        print('best score of g(S): ', self.bestAveScore)
        res3 = 'min value of the tree :  '
        for i in range(len(self.dtrees)):
            res3 += str((self.dtrees[i].getMin())) + ' '
        print(res3)
    
    
    def fastGreedyDecreasing(self):
        print('this is the cpu version of FlowScope')  
        print('start  greedy')

        self.initGreedy()

        self.bestAveScore = float('-inf')

        self.numDeleted = 0
        self.deleted = []
        self.bestNumDeleted = 0

        curall = 0
        for i in range(0, len(self.curScorelist), 2):
            curall += self.curAveScorelist[i] - self.alpha * self.curAveScorelist[i+1]
        print('initial score of g(S):', curall)
        while self.checkset():  # repeat deleting until one node set in null

            mold, idx, minidx = self.findmin()

            self.updataConnNode(mold=mold, index=idx, mindelta=minidx)
            
            s = 0
            for i in range(len(self.sets)):
                s += len(self.sets[i])
            if (s) > 0:
                for i in range(0, len(self.curAveScorelist), 2):
                    self.curAveScorelist[i] = self.curScorelist[i] / s
                    self.curAveScorelist[i+1] = self.curScorelist[i+1] / s
            else:
                print("something wrong in FlowScope")
                for i in range(0, len(self.curAveScorelist), 2):
                    self.curAveScorelist[i] = 0
                    self.curAveScorelist[i+1] = 0 
                    
            curAveScore =0
            for i in range(0, len(self.curAveScorelist), 2):
                curAveScore += self.curAveScorelist[i] - self.alpha * self.curAveScorelist[i+1]

            
            if curAveScore >= self.bestAveScore: 
                self.bestNumDeleted = self.numDeleted
                self.bestAveScore = curAveScore

        self.printres()
        

        finalsets = []
        finalsets.append(set(range(self.mcurlist[0].shape[0])))
        for i in range(len(self.mcurlist) - 1):
            finalsets.append(set(range(self.mcurlist[i].shape[1])))
        finalsets.append(set(range(self.mcurlist[-1].shape[1])))
        
        for i in range(self.bestNumDeleted):
            finalsets[self.deleted[i][0]].remove(self.deleted[i][1])
            

        return finalsets, self.bestAveScore


    def anomaly_detection(self, k:int=3, alpha:float = 0.8):
        return self.run(k=k, alpha=alpha)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   TensorData.py
@Desc    :   Structure of input file.
'''

# here put the import lib

from ..util.basicutil import set_trace
from . import STensor, DTensor
import pandas as pd
import numpy as np
from joblib import Parallel, delayed


class TensorData:
    def __init__(self, data: pd.DataFrame):
        self.data = data
        self.labels = data.columns

    def toDTensor(self, hastticks: bool = True):
        if hastticks:
            time_tensor = DTensor(self.data.iloc[:, 0])
            val_tensor = DTensor(self.data.iloc[:, 1:])
        else:
            time_tensor = None
            val_tensor = DTensor(self.data)
        return time_tensor, val_tensor

    def toSTensor(self, hasvalue: bool = True, mappers: dict = {}):
        if hasvalue:
            value = self.data.iloc[:, -1]
            attr = self.data.iloc[:, :-1]
        else:
            value = pd.Series([1] * len(self.data))
            attr = self.data

        for i in attr.columns:
            if i in mappers:
                if isinstance(i, str):
                    colind = mappers[i].map(self.data.loc[:,i])
                    attr.loc[:,i] = colind
                else:
                    colind = mappers[i].map(self.data.iloc[:, i])
                    attr.iloc[:, i] = colind
        attr = attr.astype('int')
        return STensor((attr.to_numpy().T, value.to_numpy()))
        

    def do_map(self, hasvalue=True, mappers={}):
        if hasvalue:
            value = self.data.iloc[:, -1]
            attr = self.data.iloc[:, :-1]
        else:
            value = pd.Series([1] * len(self.data))
            attr = self.data

        for i in attr.columns:
            if i in mappers:
                colind = mappers[i].map(self.data.iloc[:, i])
                attr.iloc[:, i] = colind
        return attr.to_numpy(), value.to_numpy()

    def log_to_time(self, time_col: int or list = 0, group_col: int or list = 1, val_col: int or list = None, format: str = '%Y-%m-%d %H:%M:%S', bins: int = 10, range: tuple = None, inplace: bool = False):
        """ Transfer log data to time series data.

        Parameters:
        ----------
        time_col : int or list, optional
            position of time column, default is 0

        group_col : int or list, optional
            positions of columns used to group data, default is 1

        val_col : int or list, optional
            positions of columns to be aggregated, default is None, will aggregate all but time col and group col

        format : str, optional
            time format, default is '%Y-%m-%d %H:%M:%S'

        bins : int, optional
            number of equal width bins in the given range, default is 10
        """
        import time
        _data = self.data
        if type(group_col) != list:
            group_col = [group_col]
        if val_col is not None:
            if type(val_col) != list:
                val_col = [val_col]
        else:
            val_col = [x for x in _data.columns if x not in group_col + [time_col]]
        _data = _data.iloc[:, [time_col] + group_col + val_col]
        _data.iloc[:, time_col] = _data.iloc[:, time_col].apply(lambda x: time.mktime(time.strptime(x, '%Y-%m-%d %H:%M:%S')))
        if range is not None and len(range) == 2:
            _start, _end = range
            _data = _data.iloc[_start:_end, :]
        _min = _data.iloc[:, time_col].min()
        _max = _data.iloc[:, time_col].max()
        _interval = (_max - _min) / (bins)
        _data.iloc[:, time_col] = _data.iloc[:, time_col].apply(lambda x: int((x - _min) / _interval) * _interval + _min if int((x - _min) / _interval) < bins else int((x - _min) / _interval - 1) * _interval + _min)
        name_col = [time_col] + group_col
        grouped_data = _data.groupby(name_col).sum()
        import numpy as np
        time_df = pd.DataFrame({
            time_col: np.linspace(_min, _max, bins+1)
        })
        grouped_data = grouped_data.unstack(level=group_col)
        _ans = time_df.join(grouped_data, on=time_col, how='outer')
        if inplace:
            self.data = _ans
            self.labels = list(_ans.columns)
            self.labels.remove(self.labels[time_col])
        else:
            return _ans
    
    def to_aggts(self, data, time_col: int = 0, group_col: int or list = 1, inplace: bool = False):
        aggts = {} # final dict list for aggregating time series.
        for row in data:
            if len(group_col) == 1:
                key = row[group_col[0]]
            else:
                key = ','.join(np.array(row)[group_col])
            if key not in aggts:
                aggts[key] = []
            aggts[key].append(row[time_col])
        return aggts


class TensorStream():
    def __init__(self, f, col_idx: list = None, col_types: list = None, sep: str = ' ',
                 mappers: dict = {}, hasvalue: bool = True):
        self.f = f

        if col_types is None:
            if col_idx is None:
                self.idxtypes = None
            else:
                self.idxtypes = [(x, str) for x in col_idx]
        else:
            if col_idx is None:
                col_idx = [i for i in range(len(col_types))]
            if len(col_idx) == len(col_types):
                self.idxtypes = [(x, col_types[i]) for i, x in enumerate(col_idx)]
            else:
                raise Exception(f"Error: input same size of col_types and col_idx")
        self.sep = sep
        self.mappers = mappers
        self.mappers = mappers
        self.hasvalue = hasvalue

        self.lastwindow = []  # data of last window (lines)
        self.laststrides = []  # indices of strides of last window

    def _get_file_end_pos(self):
        cur_pos = self.f.tell()
        self.f.seek(0, 2)
        end_pos = self.f.tell()
        self.f.seek(cur_pos, 0)
        return end_pos

    def fetch_slide_window(self, window: int = 10, stride: int = 5, ts_colidx: int = 0):
        end_pos = self._get_file_end_pos()
        if self.f.tell() == end_pos:
            raise Exception('all data has been processed')

        tensorlist = []
        lineid = 0

        while True:
            line = self.f.readline()
            coords = line.strip().split(self.sep)
            tline = []
            try:
                for i, tp in self.idxtypes:
                    if i == ts_colidx:
                        ts = tp(coords[i])
                        'map time during reading data'
                        if i in self.mappers:
                            ts = self.mappers[i].map([ts])[0]
                    tline.append(tp(coords[i]))
            except Exception:
                raise Exception(f"The {i}-th col does not match the given type {tp} in line:\n{line}")
            if lineid == 0:
                start_ts = ts
            else:
                if ts - start_ts >= stride:
                    if len(self.lastwindow) == 0:  # first window
                        self.laststrides.append(lineid)
                        if ts - start_ts >= window:
                            self.lastwindow = tensorlist
                            break
                    else:
                        win_start_lineid = self.laststrides.pop(0)
                        for i in range(len(self.laststrides)):
                            self.laststrides[i] -= win_start_lineid
                        curwindow = []
                        curwindow.extend(self.lastwindow[win_start_lineid:])
                        curwindow.extend(tensorlist)
                        self.lastwindow = curwindow
                        self.laststrides.append(len(curwindow))
                        break
                if self.f.tell() == end_pos:
                    break
            tensorlist.append(tline)
            lineid += 1
        tensorlist = pd.DataFrame(tensorlist)

        'map other columns, e.g. user, item'
        for i in tensorlist.columns:
            if i in self.mappers:
                colind = self.mappers[i].map(tensorlist.iloc[:, i])
                tensorlist.iloc[:, i] = colind
        return tensorlist



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py
@Desc    :   Import part and configuration part, including names exposed to spartan.
'''

# here put the import lib

from spartan.util import MODEL_PATH

from .anomaly_detection import ADPolicy, AnomalyDetection
from .forecast import ForePolicy, Forecast
from .summarization import SumPolicy, Summarization
from .train import TrainPolicy, Train

__all__ = [
    'ADPolicy', 'AnomalyDetection',
    'ForePolicy', 'Forecast',
    'SumPolicy', 'Summarization',
    'TrainPolicy', 'Train'
]



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

#  Project: eaglemine
#    __init__.py.py
#      Version:  1.0
#      Goal: Class file / Test script
#      Created by @wenchieh  on <12/17/2017>


__author__ = 'wenchieh'



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   _task.py
@Desc    :   Interface of base class Task.
'''

# here put the import lib

import importlib


class Task():
    '''Basic class of task entity.

    Bridge for data, model and parameters.

    Attributes
    ----------
    tensor : object
        data object
    policy : str
        model object
    model_name : str
        model name string
    params : dict
        parameters dictionary, used for construction

    '''

    def __init__(self, *args, **kwargs):
        '''Initialization function.'''
        self.tensor = None
        self.policy = None
        self.model_name = None
        self.model = None
        self.params = None

    @classmethod
    def create(cls, tensor: object, policy: str, model_name: str, *args, **kwargs) -> object:
        '''Create function, called by class.

        Instantiation of policy from string and creation of Task from other parameters.

        Parameters
        ----------
        tensor : object
            data object
        policy : str
            model path, defined by an enum class
        model_name : str
            model name string
        params : dict
            parameters dictionary, used for construction

        Returns
        ----------
        Task
            object for task
        '''
        try:
            model_cls = importlib.import_module(policy.value).__call__()
        except Exception as e:
            print(e)
            raise Exception(f"{policy} Not Supported!")
        model = model_cls.__create__(tensor, *args, **kwargs)
        obj = cls()
        obj.tensor = tensor
        obj.model = model
        obj.policy = policy
        obj.model_name = model_name
        obj.params = kwargs
        return obj

    def run(self, __func__: str, *args, **kwargs):
        '''Interface of run function, overrided by subclasses.

        Raises
        ----------
        NotImplementedError
            when called, return not implemented error
        '''
        raise NotImplementedError(f"{self.policy} do not implement {__func__} function.")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   timeseries.py
@Desc    :   Definition of timeseries structure.
'''

# here put the import lib
from . import DTensor


class Timeseries:
    def __init__(self, val_tensor: DTensor, time_tensor: DTensor = None, labels: list = None, freq: int = 1, startts: int = 0):
        """A class designed for time series data.

        Parameters
        ----------
        val_tensor : DTensor
            value tensor

        time_tensor : DTensor
            time tensor, default is None

        labels : list
            list of column names, default is None

        freq : int
            frequency of this series, default is 1

        startts : int
            start timetick, default is 0

        Examples
        ----------
        Timeseries can be constructed in many styles. Among all parameters, only val_tensor is necessary.

        Normally, val_tensor, time_tensor, and labels are passed in. Length of labels and val_tensor will be determined to be equal.
        And meanwhile, freq, startts will be invalid and inferred from time tensor.

        >>> Timeseries(val_tensor, time_tensor, labels=['col_1', 'col_2'])

        If labels are missing, program will defaultly assign a list of labels, as ['dim_1', 'dim_2', ...]

        >>> Timeseries(val_tensor, time_tensor)

        If time tensor is missed, program will automatically create a time tensor with parameter freq and startts.

        >>> Timeseries(val_tensor, freq=2, startts=100)
        """
        self.freq = freq
        self.val_tensor = val_tensor.T
        self.dimension, self.length = self.val_tensor.shape
        if labels is None:
            self.labels = ['dim_' + str(i) for i in range(self.dimension)]
        else:
            self.labels = list(labels)
        if time_tensor is None:
            self.startts = startts
            import numpy as np
            self.time_tensor = self.__init_time(self.val_tensor.shape[1], self.freq, self.startts)
        else:
            self.startts = time_tensor[0]
            self.freq = (self.length) / (time_tensor.max() - time_tensor.min())
            self.time_tensor = time_tensor

    def __len__(self):
        """Return and update length of time tensor as length of time series object.

        Returns
        ----------
        self.length
            length of time series object
        """
        self.length = self.time_tensor.__len__()[1]
        return self.length

    def __str__(self):
        """Return discription of time series object.

        Returns
        ----------
        _str : str
            discription of time series object
        """
        import pandas as pd
        _str = f"""
            Time Series Object
            Dimension Size: {self.dimension}
            Length: {self.length}
            Time Length: {round(self.time_tensor.max() - self.time_tensor.min(), 3)}
            Frequency: {round(self.freq, 3)}
            Start Timestamp: {round(self.startts, 3)}
            Labels: {', '.join([str(x) for x in self.labels])}
        """
        columns = ['Time']
        columns.extend(self.labels)
        print(pd.DataFrame(DTensor([self.time_tensor]).concatenate(self.val_tensor, axis=0)._data.T,
                           columns=columns))
        return _str

    def __copy__(self):
        """Return copy of time series object.

        Returns
        ----------
        object
            copy of time series object
        """
        import copy
        time_tensor = copy.copy(self.time_tensor)
        val_tensor = copy.copy(self.val_tensor).T
        labels = copy.copy(self.labels)
        return Timeseries(val_tensor, time_tensor, labels)

    def resample(self, resampled_freq: int, inplace: bool = False, show: bool = False):
        """Resample series data with a new frequency, acomplished on the basis of scipy.signal.sample.

        Parameters
        ----------
        resampled_freq : int
            resampled frequency

        inplace : bool
            update origin object or return a new object, default is False

        show : bool
            if True, draw plot

        Returns
        ----------
        None or Timeseries object
            self or a new resampled object    
        """
        _ori_tensor = DTensor.from_numpy(self.val_tensor._data.copy())
        _ori_freq = self.freq
        _self = self.__handle_inplace(inplace)
        new_len = int(_self.length / _self.freq * resampled_freq)
        _self.val_tensor.resample(new_len, inplace=True)
        _self.__update_time(_self.val_tensor, resampled_freq, _self.startts)
        _self.__update_info(_self.labels, _self.time_tensor, _self.val_tensor)
        if show:
            from spartan.util.drawutil import plot_resampled_series
            plot_resampled_series(self, self.length, _self.length, _ori_freq, resampled_freq, _ori_tensor._data, _self.val_tensor._data, _self.startts)
        if not inplace:
            return _self

    def add_columns(self, attrs: list or str, values: [int, float, DTensor, list] = None, inplace: bool = False, show: bool = False):
        """Add new equal-length columns to Time series object.

        Parameters
        ----------
        attrs : list or str
            list or string of column names

        values: [int, float, DTensor, list]
            if type of values is int or float, function will create a equal-length ndarray filled with values
            if type of values is DTensor or list, function will judge the length, then add columns
            default is None

        inplace : bool
            update origin object or return a new object, default is False

        show : bool
            if True, draw plot

        Returns
        ----------
        None or Timeseries object
            self or a new object with columns added
        """
        _self = self.__handle_inplace(inplace)
        _names_type = type(attrs)
        _values_type = type(values)
        if _names_type == str:
            if _values_type in [int, float]:
                _self.__add_single_column(attrs, values, _type='number')
            elif _values_type == DTensor:
                assert len(values.shape) == 1
                _self.__add_single_column(attrs, values, _type='tensor')
            elif _values_type == list:
                assert len(values) == 1
                _self.__add_single_column(attrs, values[0], _type='number')
            else:
                raise TypeError(f"Inappropriate values type of {type(values)}")
        elif _names_type == list:
            if _values_type == DTensor:
                assert values.shape[0] == len(attrs)
                _self.__add_multi_columns(attrs, values, _type='tensor')
            elif _values_type == list:
                assert len(values) == len(attrs)
                _value_type = type(values[0])
                if _value_type in [int, float]:
                    _self.__add_multi_columns(attrs, values, _type='number')
        _self.__handle_plot(show)
        if not inplace:
            return _self

    def __add_multi_columns(self, attrs: list, tensor: DTensor, _type: str):
        """Private function for adding multiple columns, adding operation is finished by concatenate.

        Parameters
        ----------
        attrs : list
            list of column names

        tensor : DTensor
            tensor to be added

        _type : str
            if number, function will create an equal-length ndarray for DTensor
            if tensor, function will concatenate directly
        """
        if _type == 'number':
            import numpy as np
            tensor = DTensor.from_numpy(np.tile(np.array([tensor]).T, (1, self.length)))
        elif _type == 'tensor':
            tensor = tensor
        self.val_tensor.concatenate(tensor, inplace=True)
        self.labels.extend(attrs)
        self.dimension += len(attrs)

    def __add_single_column(self, attr: str, value: DTensor, _type: str):
        """Private function for adding single column, adding operation is finished by concatenate.

        Parameters
        ----------
        columns_names : str
            string of column name

        tensor : DTensor
            tensor to be added

        _type : str
            if number, function will create an equal-length ndarray for DTensor
            if tensor, function will concatenate directly
        """
        if _type == 'number':
            import numpy as np
            _data = DTensor.from_numpy(np.array([[value] * self.length]))
        elif _type == 'tensor':
            _data = value
        self.val_tensor.concatenate(_data, inplace=True)
        self.labels.append(attr)
        self.dimension += 1

    def concat(self, series: list or "Timeseries", inplace: bool = False, show: bool = False):
        """Concatenate self with another timeseries object with the same dimension.

        Parameters
        ----------
        series : list or Timeseries
            list of Timeseries object or Timeseries object

        inplace : bool
            update origin object or return a new object, default is False

        show : bool
            if True, draw plot

        Returns
        ----------
        None or Timeseries object
            self or a new object with columns concatenated
        """
        _self = self.__handle_inplace(inplace)
        _type = type(series)
        if _type == list:
            _series = []
            for x in series:
                if type(x) == Timeseries:
                    _series.append(x.__copy__())
                else:
                    raise Exception(f'list contains non-Timeseries object')
            _self.__concat_several(_series)
        elif _type == Timeseries:
            _self.__concat_one(series.__copy__())
        _self.__update_time(_self.val_tensor, _self.freq, _self.startts)
        _self.__update_info(_self.labels, _self.time_tensor, _self.val_tensor)
        _self.__handle_plot(show)
        if not inplace:
            return _self

    def __concat_one(self, serie: "Timeseries"):
        """Private function for concating single object.

        Parameters
        ----------
        serie : Timeseries
            serie to be concatenated
        """
        if not self.dimension == serie.dimension:
            raise Exception(f'dimension sizes are not the same with self {self.dimension} and object {serie.dimension}')
        for i in range(len(self.labels)):
            if not self.labels[i] == serie.labels[i]:
                raise Exception(f'{i}th dimension is not corresponding with self {self.labels[i]} and object {serie.labels[i]}')
        self.val_tensor.concatenate(serie.val_tensor, axis=1, inplace=True)

    def __concat_several(self, concated_series: list):
        """Private function for concating several objects.

        Parameters
        ----------
        concated_series : list
            list of timeseries object to be concatenated
        """
        for serie in concated_series:
            self.__concat_one(serie)

    def combine(self, series: "Timeseries" or list, inplace: bool = False, show: bool = False):
        """Combine self with columns of other timeseries objects which have the same length.

        Parameters
        ----------
        series : list or Timeseries
            list of Timeseries object or Timeseries object

        inplace : bool
            update origin object or return a new object, default is False

        show : bool
            if True, draw plot

        Returns
        ----------
        None or Timeseries object
            self or a new object with columns combined
        """
        _self = self.__handle_inplace(inplace)
        _type = type(series)
        if _type == list:
            _series = []
            for x in series:
                if type(x) == Timeseries:
                    _series.append(x.__copy__())
                else:
                    raise Exception(f'list contains non-STTimeseries object')
            _self.__combine_several(_series)
        elif _type == Timeseries:
            _self.__combine_one(series.__copy__())
        _self.__handle_plot(show)
        if not inplace:
            return _self

    def __combine_one(self, serie: "Timeseries"):
        """Private function for combining single object.

        Parameters
        ----------
        serie : Timeseries
            serie to be combined
        """
        if not self.freq == serie.freq:
            raise Exception(f'Frequency not matched, with {self.freq} and {serie.freq}')
        for label in serie.labels:
            if label in self.labels:
                for i in range(1, 10000):
                    if not (label + '_' + str(i)) in self.labels:
                        self.labels.extend([label + '_' + str(i)])
                        break
            else:
                self.labels.extend([label])
        self.dimension += serie.dimension
        self.val_tensor.concatenate(serie.val_tensor, axis=0, inplace=True)

    def __combine_several(self, combined_series: list):
        """Private function for combining several objects.

        Parameters
        ----------
        combined_series : list
            list of timeseries object to be combined
        """
        for serie in combined_series:
            self.__combine_one(serie)

    def extract(self, attrs: list or str = None, inplace: bool = False, show: bool = False):
        """Extract specific columns from self.

        Parameters
        ----------
        attrs : list or str
            list or string of column names, default is None

        inplace : bool
            update origin object or return a new object, default is False

        show : bool
            if True, draw plot

        Returns
        ----------
        None or Timeseries object
            self or a new object with columns extracted
        """
        _self = self.__handle_inplace(inplace)
        _labels, _tensor = _self.__handle_attrs(attrs)
        _self.__update_info(_labels, _self.time_tensor, _tensor)
        _self.__handle_plot(show)
        if not inplace:
            return _self

    def cut(self, start: int = None, end: int = None, attrs: list = None, form: str = 'point', inplace: bool = False, show: bool = False):
        """Cut sub sequence from chosen attribute columns.

        Parameters
        ----------
        start : int
            start timetick or point, default is None, cut from the very front position

        end : int
            end timetick or point, default is None, cut to the very last position

        attrs : list or str
            list or string of column names, default is None, return the all columns

        form : str
            type of start and end
            if 'point', start and end stand for positions of points
            if 'time', start and end stand for timeticks of points
            default is 'point'

        inplace : bool
            update origin object or return a new object, default is False

        show : bool
            if True, draw plot

        Returns
        ----------
        None or Timeseries object
            self or a new object with tensor cut
        """
        _self = self.__handle_inplace(inplace)
        _labels, _tensor = _self.__handle_attrs(attrs)
        if form == 'point':
            start = start
            end = end
        elif form == 'time':
            if not start is None:
                start = int((start-_self.startts) * _self.freq)
            if not end is None:
                end = int((end-_self.startts) * _self.freq)
        else:
            raise Exception('Value of parameter form is not defined!')
        if start is None:
            start = 0
        if end is None:
            end = _self.length
        if start < 0 or end > _self.length:
            raise Exception(f'start pos: {start} with 0 and end pos {end} with {_self.length}')
        _self.time_tensor.cut(start, end, inplace=True)
        _tensor.cut(start, end, inplace=True)
        _self.__update_info(_labels, _self.time_tensor, _tensor)
        _self.__handle_plot(show)
        if not inplace:
            return _self

    def normalize(self, attrs: list or str = None, strategy: str = 'minmax', inplace: bool = False, show: bool = False):
        """Normalize data in value_tensor.

        Parameters
        ----------
        attrs : list or str
            list or string of column names, default is None

        strategy : str
            strategy for normalization
            if 'minmax', normalize to [-1, 1]
            default is 'minmax'

        inplace : bool
            update origin object or return a new object, default is False

        show : bool
            if True, draw plot

        Returns
        ----------
        None or Timeseries object
            self or a new object with tensor normalized
        """
        _self = self.__handle_inplace(inplace)
        _labels, _tensor = _self.__handle_attrs(attrs)
        if strategy == 'minmax':
            _tensor = _self.__normalize_minmax(_tensor)
        else:
            raise TypeError(f'strategy: {strategy} is not supported.')
        _self.__update_info(_labels, _self.time_tensor, _tensor)
        _self.__handle_plot(show)
        if not inplace:
            return _self

    def __normalize_minmax(self, _tensor: DTensor):
        """Private function for normalize value tensor by minmax function.

        Parameters
        ----------
        _tensor : DTensor
            value tensor to be normalized by minmax function

        Returns
        ----------
        _tensor : DTensor
            normalized tensor
        """
        import numpy as np
        _min = np.tile(_tensor.min(axis=1).reshape((self.dimension, 1)), self.length)
        _max = np.tile(_tensor.max(axis=1).reshape((self.dimension, 1)), self.length)
        _middle = (_min + _max) / 2
        _tensor = (_tensor - _middle) / (_max - _min) * 2
        return _tensor

    def __handle_plot(self, show: bool):
        """Private function for plotting.

        Parameters
        ----------
        show : bool
            if True, call plot function in drawutils
        """
        from spartan.util.drawutil import plot_timeseries
        if show:
            plot_timeseries(self)

    def __handle_attrs(self, attrs: str or list):
        """Private function for checking labels and tensor of column names in attrs.

        Parameters
        ----------
        attrs : list or str
            list or string of column names

        Raises
        ----------
        TypeError:
            Raise if attrs is not str or list

        Exception:
            Raise if attrs has column names which are not in self.labels

        Returns
        ----------
        _labels, _tensor : list, DTensor
            Selected labels and value tensor
        """
        if type(attrs) == str:
            attrs = [attrs]
        elif type(attrs) == list:
            attrs = attrs
        elif attrs is not None:
            raise TypeError(f'Type of attrs: {type(attrs)}')
        if attrs is None:
            _labels = self.labels
            _tensor = self.val_tensor
        else:
            _labels = []
            _tensor = []
            for attr in attrs:
                if not attr in self.labels:
                    raise Exception(f'Attr {attr} is not found')
                _labels.append(attr)
                index = self.labels.index(attr)
                _tensor.append(self.val_tensor._data[index])
            _tensor = DTensor(_tensor)
        return _labels, _tensor

    def __handle_inplace(self, inplace: bool = False):
        """Private function for checking if a new object is needed

        Parameters
        ----------
        inplace : bool
            update origin object or return a new object, default is False

        Returns
        ----------
        None or Timeseries object
            self or a new object
        """
        if inplace:
            _self = self
        else:
            import copy
            _self = copy.copy(self)
        return _self

    def __update_info(self, _labels: list, _time: DTensor, _tensor: DTensor):
        """Update infomation of self from newly updated tensors.

        Parameters
        ----------
        _labels : list
            list of column names

        _time : DTensor
            time tensor

        _tensor : DTensor
            value tensor
        """
        assert len(_labels) == len(_tensor)
        self.labels, self.time_tensor = _labels, _time
        self.val_tensor, self.dimension = _tensor, len(_tensor)
        self.startts = self.time_tensor[0]
        self.length = self.val_tensor.shape[1]

    def __update_time(self, val_tensor: DTensor, freq: int, startts: int):
        """Update infomation of self from newly updated tensors.

        Parameters
        ----------
        val_tensor : DTensor
            value tensor

        freq : int
            frequency of series

        startts : int
            start time tick
        """
        _len = val_tensor.shape[1]
        self.length = _len
        self.time_tensor = self.__init_time(_len, freq, startts)
        self.freq = freq
    
    
    def __init_time(self, len: int, freq: int, startts: int):
        """Construct time tensor.

        Parameters
        ----------
        len : int
            length of time tensor

        freq : int
            frequency of series

        startts : int
            start time tick
        
        Returns
        ----------
        time_tensor : DTensor
            time tensor
        """
        import numpy as np
        time_tensor = DTensor.from_numpy(np.linspace(startts, 1 / freq * len + startts - 1, len))
        return time_tensor



#!/usr/bin/python
# -*- coding=utf-8 -*-

#  Project: eaglemine
#  File: eaglemine.py
#  Goal: The main routine of eaglemine
#      Version:  1.0
#      Created by @wenchieh  on <12/17/2017>
#

__author__ = 'wenchieh'

# sys
import time

# third-party lib
import numpy as np
from scipy.sparse.linalg import svds

# project
from .eaglemine_model import EagleMineModel
from .tools.histogram_heuristic_generator import HistogramHeuristicGenerator

from .._model import DMmodel
from spartan.tensor.graph import Graph


class EagleMine( DMmodel ):
    ''' Micro-cluster detection: vision-guided anomaly detection.
    Given a histogram derived from the correlated features of graph nodes，
    EagleMine can be used to identify the micro-clusters in the graph,
    these nodes in micro-clusters basically corresponds to some anomaly patterns.
    '''
    def __init__(self, voctype:str="dtmnorm", mode:int=2, mix_comps:int=2):
        '''Initialization func for EagleMine
        Parameters:
        --------
        :param voctype: str
            vocabulary type: {"dtmnorm", "dmgauss"}
            Default is "dtmnorm".
        :param mode: int
            The dimensions of features (the histogram).
            Default is 2.
        :param mix_comps: int
            # mixture component for describing the major island.
            Default is 2.
        '''
        self.voctype = voctype
        self.mode = mode
        self.ncomps = mix_comps
        self.eaglemodel = None
        self.__histgenerator__ = None
    
    @classmethod
    def __create__(cls, *args, **kwargs):
        return cls(*args, **kwargs)
    
    @staticmethod
    def get_graph_spectral(sparse_matrix):
        '''Extract the spectral features of the given sparse matrix (graph)
        Parameter:
        --------
        :param sparse_matrix:
            sparse matrix for adjacency matrix a graph.
        '''
        hub, _, auth = svds(1.0 * sparse_matrix.tocsr(), k=1, which='LM')
        hub, auth = np.squeeze(np.array(hub)), np.squeeze(np.array(auth))
        if abs(np.max(hub)) < abs(np.min(hub)):
            hub *= -1
        hub[hub < 0] = 0
        if abs(np.max(auth)) < abs(np.min(auth)):
            auth *= -1
        auth[auth < 0] = 0
        
        return hub, auth
    
    def graph2feature(self, graph:Graph, feature_type:str='outdegree2hubness'):
        '''Extract example correlated-features of the given graph and generate the corresponding histogram.
        Parameters:
        --------
        :param graph: Graph
            Given graph data
        :param feature_type: str
            Feature type for the graph node: {'outdegree2hubness', 'indegree2authority', 'degree2pagerank'}.
            Default is 'outdegree2hubness'.
        '''
        
        if feature_type not in {'outdegree2hubness', 'indegree2authority', 'degree2pagerank'}:
            raise NameError("Invalid feature type 'ty_type', which should be in {'outdegree2hubness', 'indegree2authority', 'degree2pagerank'}")
        
        print("extract graph features ..... ")
        start_tm = time.time()
        outd, ind = np.asarray(graph.sm.sum(axis=1)).squeeze(), np.asarray(graph.sm.sum(axis=0)).squeeze()
        hub, auth = self.get_graph_spectral(graph.sm)
        if feature_type == 'outdegree2hubness' or feature_type == 'degree2pagerank':
            deg, spec = outd, hub
        else:
            deg, spec = ind, auth
        degreeidx = 0
        print("done! @ {}s".format(time.time() - start_tm))
        
        return degreeidx, np.column_stack((deg, spec))
    
    def feature2histogram(self, feature:np.ndarray, degreeidx:int=0, 
                          N_bins:int=80, base:int=10, mode:int=2, verbose:bool=True):
        '''Construct histogram with given features
        Parameters:
        --------
        :param feature: np.ndarray
            The correlated features
        :param degreeidx: int
            The index of 'degree' ('out-degree', 'in-degree') in features, degreeidx=-1 if not containing 'degree' feature
            Default is 0.
        :param N_bins: int
            The expected number of bins for generating histogram.
            Default is 80.
        :param base: int
            The logarithmic base for bucketing the graph features.
            Default is 10.
        :param mode: int
            The dimensions of features for constructing the histogram.
            Default is 2.
        :param verbose: bool
            Whether output some running logs.
            Default is True.
        '''
        n_samples, n_features = feature.shape
        index = np.array([True] * n_samples)
        for mod in range(mode):
            index &= feature[:, mod] > 0
        if verbose:
            print("total shape: {}, valid samples:{}".format(feature.shape, np.sum(index)))
    
        degree, feat = None, None
        feature = feature[index, :]
        if degreeidx >= 0:
            degree = feature[:, degreeidx]
            feat = np.delete(feature, degreeidx, axis=1)
            del feature
        else:
            feat = feature
        
        print("construct histogram ..... ")
        start_tm = time.time()
        self.__histgenerator__ = HistogramHeuristicGenerator()
        if degree is not None:
            self.__histgenerator__.set_deg_data(degree, feat)
            self.__histgenerator__.histogram_gen(method="degree", N=N_bins, base=base)
        else:
            self.__histgenerator__.set_data(feat)
            self.__histgenerator__.histogram_gen(method="N", N=N_bins, logarithmic=True, base=base)
        print("done! @ {}s".format(time.time() - start_tm))
        
        if verbose:
            self.__histgenerator__.dump()
        
        n_nodes = len(self.__histgenerator__.points_coord)
        node2hcel = map(tuple, self.__histgenerator__.points_coord)
        nodeidx2hcel = dict(zip(range(n_nodes), node2hcel))
        
        return self.__histgenerator__.histogram, nodeidx2hcel, self.__histgenerator__.hpos2avgfeat
    
    def set_histdata(self, histogram:dict, node2hcel:dict, hcel2avgfeat:dict, weighted_ftidx:int=0):
        '''Set the histogram data
        Parameters:
        --------
        :param histogram: dict
            the format '(x,y,z,...): val', denoting that the cell (x,y,z,...) affiliates with value 'val'.
        :param node2hcel: dict
            graph node id (index) to histogram cell
        :param hcel2avgfeat: dict
            the average feature values for each histogram cell.
        :param weighted_ftidx: int
            The feature index as weight for suspiciousness metric.
            Default is 0.
        '''
        self.histogram, self.hcelsusp_wt = list(), list()
        
        for hcel in histogram.keys():
            self.histogram.append(list(hcel) + [histogram.get(hcel)])
            self.hcelsusp_wt.append(hcel2avgfeat[hcel][weighted_ftidx])
        
        self.histogram = np.asarray(self.histogram)
        self.hcelsusp_wt = np.asarray(self.hcelsusp_wt)
        self.node2hcel = np.column_stack((list(node2hcel.keys()), list(node2hcel.values())))
        self.eaglemodel = None
        self.__histgenerator__ = None
        self.hcel2label = None
        
    def run(self, outs:str, waterlevel_step:float=0.2, prune_alpha:float=0.80, 
            min_pts:int=20, strictness:int=3, verbose:bool=True):
        ''' micro-cluster identification and refinement with water-level tree.
        Parameters:
        --------
        :param outs: str
            Output path for some temporary results.
        :param waterlevel_step: float
            Step size for raising the water level.
            Default is 0.2.
        :param prune_alpha: float
            How proportion of pruning for level-tree.
            Default is 0.80.
        :param min_pts: int
            The minimum number of points in a histogram cell.
            Default is 20.
        :param strictness: int
            How strict should the anderson-darling test for normality. 0: not at all strict; 4: very strict
            Default is 3.
        :param verbose: bool
            Whether output some running logs.
            Default is True.
        '''
        print("*****************")
        print("[0]. initialization")
        self.eaglemodel = EagleMineModel(self.mode, self.ncomps)
        self.eaglemodel.set_vocabulary(self.voctype)
        self.eaglemodel.set_histogram(self.histogram)
        
        print("*****************")
        print("[1]. WaterLevelTree")
        start_tm = time.time()
        self.eaglemodel.leveltree_build(outs, waterlevel_step, prune_alpha, verbose=verbose)
        end_tm1 = time.time()
        print("done @ {}".format(end_tm1 - start_tm))
        
        print("*****************")
        print("[2]. TreeExplore")
        self.eaglemodel.search(min_pts, strictness, verbose=verbose)
        self.eaglemodel.post_stitch(strictness, verbose=verbose)
        end_tm2 = time.time()
        print("done @ {}".format(end_tm2 - end_tm1))
        
        print("*****************")
        print("[3]. node groups cluster and suspicious measure")
        self.hcel2label, mdl = self.eaglemodel.cluster_remarks(strictness, verbose=verbose)
        cluster_suspicious = self.eaglemodel.cluster_weighted_suspicious(self.hcelsusp_wt, strictness, verbose=verbose)
        
        # print("description length (mdl): {}".format(mdl))
        # print("suspicious result: {}".foramt(cluster_suspicious))
        
        print('done @ {}'.format(time.time() - start_tm))
    
    def __str__(self):
        return str(vars(self))

    def dump(self):
        self.eaglemodel.dump()
        if self.__histgenerator__ is not None:
            self.__histgenerator__.dump()
        print("done!")

    def save(self, outfn_eaglemine:str, outfn_leveltree:str=None, 
             outfn_node2label:str=None, outfn_hcel2label:str=None, 
             comments:str="#", delimiter:str=";"):
        '''save result of EagleMine
        Parameters:
        --------
        :param outfn_eaglemine: str
            Output path for eaglemine data
        :param outfn_leveltree: str
            Output path for the eater-level-tree data.
        :param outfn_node2label: str
            Output path for node2label data
        :param outfn_hcel2label: str
            Output path for hcel2label data
        :param comments: str
            The comments (start character) of inputs.
            Default is "#".
        :param delimiter: str
            The separator of items in each line of inputs.
            Default is ";".
        '''
        print("saving result")
        start_tm = time.time()
        self.eaglemodel.save(outfn_eaglemine)
        if outfn_leveltree:
            self.eaglemodel.leveltree.save_leveltree(outfn_leveltree, verbose=False)
        
        nlabs = len(np.unique(list(self.hcel2label.values())))
        if outfn_node2label is not None:
            nnds = len(self.node2hcel)
            with open(outfn_node2label, 'w') as ofp_node2lab:
                ofp_node2lab.writelines(comments + " #pt: {}, #label: {}\n".format(nnds, nlabs))
                for k in range(nnds):
                    nodeidx, hcel = self.node2hcel[k, 0], tuple(self.node2hcel[k, 1:])
                    nodelab = self.hcel2label.get(hcel, -1)
                    ofp_node2lab.writelines("{}{}{}\n".format(nodeidx, delimiter, nodelab))
                ofp_node2lab.close()
                
        if outfn_hcel2label is not None:
            nhcels = len(self.hcelsusp_wt)
            with open(outfn_hcel2label, 'w') as ofp_hcel2lab:
                ofp_hcel2lab.writelines(comments + ' #hcel: {}, #label: {}'.format(nhcels, nlabs))
                for hcel, lab in self.hcel2label.items():
                    hcel_str = delimiter.join(map(str, hcel))
                    ofp_hcel2lab.writelines("{}{}{}\n".format(hcel_str, delimiter, lab))
                ofp_hcel2lab.close()

        end_tm = time.time()        
        print("done! @ {}s".format(end_tm - start_tm))
    
    def save_histogram(self,outfn_histogram:str=None, outfn_node2hcel:str=None, outfn_hcel2avgfeat:str=None, 
                            comments:str="#", delimiter:str=","):
        '''Save the histogram data for the graph.
        Parameters:
        --------
        :param outfn_histogram: str
            Output path of histogram.
            The record in histogram should be in the format 'x,y,z,...,val', denoting that the cell (x, y, z, ...) affiliates with value 'val'
            Default is None.
        :param outfn_node2hcel: str
            Output path of the file mapping the node to histogram cell.
            Default is None.
        :param outfn_hcel2avgfeat: str
            Output path of the file mapping the histogram cell to the average features and #points
            Default is None.
        :param comments: str
            The comments (start character) of inputs.
            Default is "#".
        :param delimiter: str
            The separator of items in each line of inputs.
            Default is ",".
        '''
        if self.__histgenerator__ is not None:
            if outfn_histogram is not None:
                self.__histgenerator__.save_histogram(outfn_histogram, delimiter, comments)
            if outfn_node2hcel is not None:
                self.__histgenerator__.save_pts_index(outfn_node2hcel, delimiter, comments)
            if outfn_hcel2avgfeat is not None:
                self.__histgenerator__.save_hpos2avgfeat(outfn_hcel2avgfeat, delimiter, comments)
        else:
            RuntimeError("No histogram generated for given graph!")
    
    ## TODO: relized the 'anomaly_detection' for the task based running. The root API may need to be refactored.
    # def anomaly_detection(self, outs:str, waterlevel_step:float=0.2, 
    #                       prune_alpha:float=0.80, min_pts:int=20, strictness:int=3):
    #     '''anomaly detection with EagleMine
    #     Parameters:
    #     --------
    #     outs: str
    #         Output path for some temporary results.
    #     waterlevel_step: float
    #         Step size for raising the water level.
    #         Default is 0.2.
    #     prune_alpha: float
    #         How proportion of pruning for level-tree.
    #         Default is 0.80.
    #     min_pts: int
    #         The minimum number of points in a histogram cell.
    #         Default is 20.
    #     strictness: int
    #         How strict should the anderson-darling test for normality. 0: not at all strict; 4: very strict
    #         Default is 3.
    #     '''
    #     return self.run(outs, waterlevel_step, prune_alpha, min_pts, strictness, True)
    



import sys,math
import numpy as np
import scipy.sparse.linalg as slin
from scipy.sparse import coo_matrix, csr_matrix, csc_matrix
from svddenseblock import *
from mytools.ioutil import myreadfile
from os.path import expanduser
home = expanduser("~")


def loadtensor2matricization(tensorfile, sumout=[], mtype=coo_matrix,
                             weighted=True, dtype=int):
    'sumout: marginized (sumout) the given ways'
    matcols={}
    rindexcols={}
    xs, ys, data = [], [], []
    with myreadfile(tensorfile, 'rb') as f:
        for line in f:
            elems = line.strip().split(',')
            elems = np.array(elems)
            u = int(elems[0])
            colidx = range(1,len(elems)-1) #remove sumout
            colidx = set(colidx) - set(list(sumout))
            colidx = sorted(list(colidx))
            col=' '.join(elems[colidx])
            if col not in matcols:
                idx = len(matcols)
                matcols[col] = idx
                rindexcols[idx]=col
            cid = matcols[col]
            w = dtype(elems[-1])
            xs.append(u)
            ys.append(cid)
            data.append(w)
        nrow, ncol = max(xs)+1, max(ys)+1
        sm = mtype( (data, (xs, ys)), shape=(nrow, ncol), dtype=dtype )
        if weighted is False:
            sm.data[0:] = dtype(1)
        f.close()

    return sm, rindexcols

def matricizeSVDdenseblock(sm, rindexcols, rbd='avg'):
    A, tmpB = svddenseblock(sm, rbd=rbd)
    rows = A.nonzero()[0]
    cols = tmpB.nonzero()[0]
    bcols = set()
    for col in cols:
        'col name'
        cnm = rindexcols[col]
        cnm = cnm.strip().split(' ')
        b = int(cnm[0])
        bcols.add(b)
    return set(rows), set(bcols)

if __name__=="__main__":
    path = home+'/Data/BeerAdvocate/'
    respath= path+'results/'
    tsfile = path+'userbeerts.dict'
    ratefile = path+'userbeerrate.dict'
    tensorfile =respath+'userbeer.tensor'
    sm, rindexcols = loadtensor2matricization(tensorfile,
                                              sumout=[3],mtype=csr_matrix,
                                              dtype=float,weighted=True)
    A, B = matricizeSVDdenseblock(sm, rindexcols, rbd='avg')




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py
@Desc    :   Configure names exposed to spartan.
'''

# here put the import lib

from .backend import load_backend, get_backend
from .task import *
from .util import *
from .model import *
from .tensor import *



# A tree data structure which stores a list of degrees and can quickly retrieve the min degree element,
# or modify any of the degrees, each in logarithmic time. It works by creating a binary tree with the
# given elements in the leaves, where each internal node stores the min of its two children.
# min tree, which can fetch min value quickly
import math
class MinTree:
    def __init__(self, degrees):
        degrees = self.checktype(degrees)
        self.height = int(math.ceil(math.log(len(degrees), 2)))
        self.numLeaves = 2 ** self.height  #  the operater form of pow
        self.numBranches = self.numLeaves - 1
        self.n = self.numBranches + self.numLeaves
        self.nodes = [float('inf')] * self.n
        for i in range(len(degrees)):
            self.nodes[self.numBranches + i] = degrees[i]
        for i in reversed(range(self.numBranches)):
            self.nodes[i] = min(self.nodes[2 * i + 1], self.nodes[2 * i + 2])

    def getMin(self):
        cur = 0
        for i in range(self.height):
            cur = (2 * cur + 1) if self.nodes[2 * cur + 1] <= self.nodes[2 * cur + 2] else (2 * cur + 2)  #  select one from left child and right child
        return (cur - self.numBranches, self.nodes[cur])

    def index_of(self ,idx):
        idx = self.checktype(idx)
        cur = self.numBranches + idx
        return self.nodes[cur]

    # return the renewed value of degrees[idx]
    def changeVal(self, idx, delta):
        idx = self.checktype(idx)
        delta = self.checktype(delta)
        cur = self.numBranches + idx
        if self.nodes[cur] == float('inf'):
            return float('inf')

        self.nodes[cur] = self.nodes[cur] + delta

        for i in range(self.height):
            cur = (cur - 1) // 2
            nextParent = min(self.nodes[2 * cur + 1], self.nodes[2 * cur + 2])
            if self.nodes[cur] == nextParent:
                break
            self.nodes[cur] = nextParent
        return self.nodes[cur]

    # return the renewed value of degrees[idx]
    def setVal (self, idx, new_value):
        idx = self.checktype(idx)
        new_value = self.checktype(new_value)
        cur = self.numBranches + idx
        if self.nodes[cur] == float('inf'):
            return float('inf')

        self.nodes[cur] = new_value
        for i in range(self.height):
            cur = (cur - 1) // 2
            nextParent = min(self.nodes[2 * cur + 1], self.nodes[2 * cur + 2])
            if self.nodes[cur] == nextParent:
                break
            self.nodes[cur] = nextParent
        return self.nodes[cur]

    def dump(self):
        print("numLeaves: %d, numBranches: %d, n: %d, nodes: " % (self.numLeaves, self.numBranches, self.n))
        cur = 0
        for i in range(self.height + 1):
            for j in range(2 ** i):
                print(self.nodes[cur],)
                cur += 1
            print('')

    def checktype(self, obj):
        obj_type = type(obj)
        if obj_type is np.ndarray:
            obj = obj.tolist()
            return obj
        elif isinstance(obj, list):
            obj = (np.array(obj)).tolist()
            return obj
        elif obj_type in [np.int_, np.intc, np.int8, np.int16, np.int32, np.int64]:
            obj = int(obj)
            return obj
        elif obj_type in [np.float_, np.float16, np.float32, np.float64]:
            obj = float(obj)
            return obj
        elif isinstance(obj, int):
            return obj
        elif isinstance(obj, float):
            return obj
        else: raise Exception('The type of ', obj,' is the wrong type!')




#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

#  Project: eaglemine
#    graph2histogram.py
#      Version:  1.0
#      Goal: Routine scripts
#      Created by @wenchieh  on <12/27/2018>


__author__ = 'wenchieh'


import numpy as np
from scipy.sparse import csr_matrix

from .utils.loader import Loader
from .tools.histogram_heuristic_generator import HistogramHeuristicGenerator
from .utils.ploter import plot_heatmap, plot_heatmap_graphlab_pgrk, plot_heatmap_2discretes


VERBOSE = True

def histogram_construct(graph_feature, degree_index, outfn_histogram, outfn_pts2pos, outfn_hpos2avgfeat, mode=2):
    n_samples, n_features = graph_feature.shape
    index = np.array([True] * n_samples)
    for mod in range(mode):
        index &= graph_feature[:, mod] > 0
    if VERBOSE:
        print("total shape: {}, valid samples:{}".format(graph_feature.shape, np.sum(index)))

    degree, features = None, None
    graph_feature = graph_feature[index, :]
    if degree_index > 0:
        degree = graph_feature[:,degree_index-1]
        features = np.delete(graph_feature, degree_index-1, axis=1)
        del graph_feature
    else:
        features = graph_feature

    hist_gen = HistogramHeuristicGenerator()
    if degree is not None:
        hist_gen.set_deg_data(degree, features)
        hist_gen.histogram_gen(modeth="degree", N=80, base=10)
    else:
        n_buckets, logarithmic, base = 100, True, 10
        hist_gen.set_data(features)
        hist_gen.histogram_gen(method="N", N=n_buckets, logarithmic=logarithmic, base=base)


    if VERBOSE: hist_gen.dump()
    hist_gen.save_histogram(outfn_histogram)
    hist_gen.save_pts_index(outfn_pts2pos, pts_idx=np.arange(n_samples)[index])
    hist_gen.save_hpos2avgfeat(outfn_hpos2avgfeat)
    print("Graph feature to histogram done!")

def histogram_view(histogram_infn, xlabel, ylabel, outfn=None):
    loader = Loader()
    _shape_, ticks_vec, hist_arr = loader.load_multi_histogram(histogram_infn)
    csr_mat = csr_matrix((hist_arr[:, -1], (hist_arr[:, 0], hist_arr[:, 1])), shape=_shape_, dtype=int)
    plot_heatmap(ticks_vec[1], ticks_vec[0], csr_mat.toarray(), xlabel=xlabel, ylabel=ylabel, outfn=outfn)
    # plot_heatmap_graphlab_pgrk(ticks_vec[1], ticks_vec[0], csr_mat.toarray(), xlabel=xlabel, ylabel=ylabel, outfn=outfn)
    # plot_heatmap_2discretes(ticks_vec[1], ticks_vec[0], csr_mat.toarray(), xlabel=xlabel, ylabel=ylabel, outfn=outfn)
    print('Histogram view done!')


if __name__ == '__main__':
    ins_gfeat = '../example/outd2hub_feature'
    outs = '../output/'
    outs_hist = 'histogram.out'
    ofn_node2hcel = 'node2hcel.out'
    ofn_hcel2avgfeat = 'hcel2avgfeat.out'
    ofn_heatmap = 'heatmap.png'
    x_lab, y_lab = ["Hubness", "Out-degree"]
              # ["Authoritativeness", "In-degree"], ["PageRank", "Degree"], ["Degree", "Triangles"]

    mode = 2
    loader = Loader()
    m, _, gfts = loader.load_features(ins_gfeat, float)
    histogram_construct(gfts[:m], 1, outs + outs_hist, outs + ofn_node2hcel, outs + ofn_hcel2avgfeat, mode)
    histogram_view(outs + outs_hist, x_lab, y_lab, outs + ofn_heatmap)




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   main.py
@Desc    :   Test entry point.
'''

# here put the import lib

import unittest
import os

s = unittest.TestSuite()
loader = unittest.TestLoader()
s.addTests(loader.discover(os.getcwd()))
run = unittest.TextTestRunner()
run.run(s)



import numpy as np
from .._model import Generalmodel
# import spartan2.ioutil as ioutil
from spartan.util.ioutil import saveDictListData, loadDictListData


class IAT(Generalmodel):
    aggiat = {}  # key:user; value:iat list
    user_iatpair = {}  # key:user; value: (iat1, iat2) list
    iatpair_user = {}  # key:(iat1, iat2) list; value: user
    iatpaircount = {}  # key:(iat1, iat2); value:count
    iatcount = {}  # key:iat; value:count
    iatprob = {} # key:iat; value:probability
    usrdict = {} # key:usr, value:frequency

    def __init__(self, aggiat={}, user_iatpair={}, iatpair_user={}, iatpaircount={}, iatcount={}):
        self.aggiat = aggiat
        self.user_iatpair = user_iatpair
        self.iatpair_user = iatpair_user
        self.iatpaircount = iatpaircount
        self.iatcount = iatcount

    def calaggiat(self, aggts):
        'aggts: key->user; value->timestamp list'
        for k, lst in aggts.items():
            if len(lst) < 2:
                continue
            lst.sort()
            iat = np.diff(lst)
            self.aggiat[k] = iat

    def save_aggiat(self, outfile):
        saveDictListData(self.aggiat, outfile)

    def load_aggiat(self, infile):
        self.aggiat = loadDictListData(infile, ktype=int, vtype=int)
    
    def get_iatpair_user_dict(self):
        'construct dict for iat pair to keys'
        for k, lst in self.aggiat.items():
            for i in range(len(lst) - 1):
                pair = (lst[i], lst[i + 1])
                if pair not in self.iatpair_user:
                    self.iatpair_user[pair] = []
                self.iatpair_user[pair].append(k)

    def get_user_iatpair_dict(self):
        for k, lst in self.aggiat.items():
            pairs = []
            for i in range(len(lst) - 1):
                pair = (lst[i], lst[i + 1])
                pairs.append(pair)
            self.user_iatpair[k] = pairs

    def getiatpairs(self):
        xs, ys = [], []
        for k, lst in self.aggiat.items():
            for i in range(len(lst) - 1):
                xs.append(lst[i])
                ys.append(lst[i + 1])
        return xs, ys

    def caliatcount(self):
        for k, lst in self.aggiat.items():
            for iat in lst:
                if iat not in self.iatcount:
                    self.iatcount[iat] = 0
                self.iatcount[iat] += 1

        allcount = sum(self.iatcount.values()) # total sum of iat
        self.iatprob = {iat: self.iatcount[iat]/allcount for iat in self.iatcount.keys()} #cal probability of iat

    def caliatpaircount(self):
        for k, lst in self.aggiat.items():
            for i in range(len(lst) - 1):
                pair = (lst[i], lst[i+1])
                if pair not in self.iatcount:
                    self.iatpaircount[pair] = 0
                self.iatpaircount[pair] += 1

    def find_iatpair_user(self, iatpairs):
        'find users that have pairs in iatpairs'
        usrset = set()
        for pair in iatpairs:
            if pair in self.iatpair_user:
                usrlist = self.iatpair_user[pair]
                usrset.update(usrlist)
        return list(usrset)
    
    def get_user_dict(self, iatpairs):
        '''get users dict that have pairs in iatpairs ordered by decreasing frequency
        Parameters:
        --------
        :param iatpairs: dict
            iat pair returned by find_peak_rect function in RectHistogram class
        '''
        for pair in iatpairs:
            if pair in self.iatpair_user:
                usrlist = self.iatpair_user[pair]
                for usr in usrlist:
                    if usr in self.usrdict:
                        self.usrdict[usr] += 1
                    else:
                        self.usrdict[usr] = 1
        self.usrdict = sorted(self.usrdict.items(), key=lambda item:item[1], reverse=True)
    
    def find_topk_user(self, k=-1):
        '''find Top-K users that have pairs in iatpairs ordered by decreasing frequency
        Parameters:
        --------
        :param k: int
            default: -1 , means return all user
            else return Top-k user
        '''
        usrlist = [usrcountpair[0] for usrcountpair in self.usrdict[:k]] 
        return usrlist
        
    def drawIatPdf(self, usrlist: list, outfig=None):
        '''Plot Iat-Pdf line
        Parameters:
        --------
        :param usrlist: list
            Top-k user returned by find_iatpair_user_ordered function
        :param outfig: str
            fig save path
        '''
        iatset = set()
        for usrid in usrlist:
            iatset.update(self.aggiat[usrid])
        iatlist = list(iatset)
        iatlist.sort()

        import matplotlib.pyplot as plt
        fig = plt.figure()
        xs = iatlist
        ys = [self.iatprob[iat] for iat in iatlist]
        plt.plot(xs, ys, 'b')
        plt.xscale('log')
        plt.xlabel('IAT(seconds)')
        plt.ylabel('pdf')
        if outfig is not None:
            fig.savefig(outfig)
        return fig
        








import matplotlib.pyplot as plt
import matplotlib
import numpy as np

class RectHistogram:
    '''
       xscale: [ 'linear' | 'log' ]
               Use a linear or log10 scale on the horizontal axis.
       yscale: [ 'linear' | 'log' ]
               Use a linear or log10 scale on the vertical axis.
       gridsize: [ 100 | integer | tuple ]
               The number of rectangles in the x-direction, default is 100. The
               corresponding number of rectangles in the y-direction is chosen such that
               the rectangles are square. Alternatively, gridsize can be
               a tuple with two elements specifying the number of rectangles in the
               x-direction and the y-direction.
        H: 2D array
        The bi-dimensional histogram of samples x and y. 
        xedges: 1D array
        The bin edges along the x axis.
        yedges: 1D array
        The bin edges along the y axis.
       '''
    xscale = 'log'
    yscale = 'log'
    gridsize = 200
    H = []
    xedges = []
    yedges = []
    xs = []
    ys = []

    def __init__(self, xscale='log', yscale='log', gridsize=200):
        self.xscale = xscale
        self.yscale = yscale
        self.gridsize = gridsize

    def draw(self, xs, ys, outfig=None, colorscale=True,
             suptitle='Rectangle binning points', xlabel='', ylabel=''):
        xs = np.array(xs) if type(xs) is list else xs
        ys = np.array(ys) if type(ys) is list else ys
        self.xs = xs
        self.ys = ys

        if self.xscale == 'log' and min(xs) <= 0:
            print('[Warning] logscale with nonpositive values in x coord')
            print('\tremove {} nonpositives'.format(len(np.argwhere(xs <= 0))))
            xg0 = xs > 0
            xs = xs[xg0]
            ys = ys[xg0]
        if self.yscale == 'log' and min(ys) <= 0:
            print('[Warning] logscale with nonpositive values in y coord')
            print('\tremove {} nonpositives'.format(len(np.argwhere(ys <= 0))))
            yg0 = ys > 0
            xs = xs[yg0]
            ys = ys[yg0]

        fig = plt.figure()

        # color scale
        cnorm = matplotlib.colors.LogNorm() if colorscale else matplotlib.colors.Normalize()
        suptitle = suptitle + ' with a log color scale' if colorscale else suptitle

        # axis space
        if isinstance(self.gridsize, tuple):
            xgridsize = self.gridsize[0]
            ygridsize = self.gridsize[1]
        else:
            xgridsize = ygridsize = self.gridsize
        if self.xscale == 'log':
            xlogmax = np.ceil(np.log10(max(xs)))
            x_space = np.logspace(0, xlogmax, xgridsize)
        else:
            x_space = xgridsize
        if self.yscale == 'log':
            ylogmax = np.ceil(np.log10(max(ys)))
            y_space = np.logspace(0, ylogmax, ygridsize)
        else:
            y_space = ygridsize
        '''
        H: 2D array
        The bi-dimensional histogram of samples x and y. 
        xedges: 1D array
        The bin edges along the x axis.
        yedges: 1D array
        The bin edges along the y axis.
        '''
        H, xedges, yedges, _ = plt.hist2d(xs, ys, bins=(x_space, y_space), cmin=1, norm=cnorm, cmap=plt.cm.jet)
        
        self.H = H
        self.xedges = xedges
        self.yedges = yedges
        
        plt.xscale(self.xscale)
        plt.yscale(self.yscale)

        plt.title(suptitle)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)

        cb = plt.colorbar()
        if colorscale:
            cb.set_label('log10(N)')
        else:
            cb.set_label('counts')

        if outfig is not None:
            fig.savefig(outfig)
        return fig
    
    def find_peak_range(self, x, y, radius):
        '''
        return: the range of coordinates which bin with the largest number of samples in the range of
                horizontal axis: [x-radius, x+radius]
                vertical axis: [y-radius, y+radius]
        '''
        xs = self.xs
        ys = self.ys
        H = self.H
        xedges = self.xedges
        yedges = self.yedges

        H[np.isnan(H)] = 0

        if self.xscale == 'log' and min(xs) <= 0:
            print('[Warning] logscale with nonpositive values in x coord')
            print('\tremove {} nonpositives'.format(len(np.argwhere(xs <= 0))))
            xg0 = xs > 0
            xs = xs[xg0]
            ys = ys[xg0]
        if self.yscale == 'log' and min(ys) <= 0:
            print('[Warning] logscale with nonpositive values in y coord')
            print('\tremove {} nonpositives'.format(len(np.argwhere(ys <= 0))))
            yg0 = ys > 0
            xs = xs[yg0]
            ys = ys[yg0]
        xmin, xmax = min(xs), max(xs)
        ymin, ymax = min(ys), max(ys)
        '''
            x range: [x - radius, x + radius]
            y range: [y - radius, y + radius]
        '''
        if x - radius < xmin:
            xst = xmin
        else:
            xst = x - radius
        if x + radius > xmax:
            xend = xmax
        else:
            xend = x + radius
        if y - radius < ymin:
            yst = ymin
        else:
            yst = y - radius
        if y + radius > ymax:
            yend = ymax
        else:
            yend = y + radius

        'find the bins in the range'
        xstpoint = np.argwhere(xedges >= xst)[0][0]
        xendpoint = np.argwhere(xedges <= xend)[-1][0]
        xbins = range(xstpoint, xendpoint)
        ystpoint = np.argwhere(yedges >= yst)[0][0]
        yendpoint = np.argwhere(yedges <= yend)[-1][0]
        ybins = range(ystpoint, yendpoint)
        'find bin with the largest number of samples in the range'
        H1 = H[xbins]
        H2 = H1[:, ybins]
        maxpoint = np.argmax(H2)
        'the index of the largest number of array H2'
        xindex, yindex = int(maxpoint / H2.shape[1]), int(maxpoint % H2.shape[1])
        'the index of bin with maximum number'
        xbinid, ybinid = xindex + xstpoint, yindex + ystpoint

        'the range of max bin'
        binxst, binxend, binyst, binyend = xedges[xbinid], xedges[xbinid + 1], \
                                           yedges[ybinid], yedges[ybinid + 1]
        return (binxst, binxend), (binyst, binyend)
    
    
    
    def find_peak_rect(self, binx: tuple, biny: tuple):
        '''
        binx: the range of max bin along the x axis.
        biny: the range of max bin along the y axis.
        return coordinate pairs in the max bin
        '''
        xs = self.xs
        ys = self.ys
        (binxst, binxend) = binx
        (binyst, binyend) = biny
        xcoords = set(np.argwhere(xs >= binxst).T[0]) & set(np.argwhere(xs <= binxend).T[0])
        ycoords = set(np.argwhere(ys >= binyst).T[0]) & set(np.argwhere(ys <= binyend).T[0])
        pairids = list(xcoords & ycoords)
        coordpairs = list(zip(xs[pairids], ys[pairids]))
        return coordpairs


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   test_ioutil.py
@Desc    :   Test file for ioutil.
'''

# here put the import lib

import unittest

from . import loadTensor, TensorData


class TestLoadTensor(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("Start Test")

    @classmethod
    def tearDownClass(cls):
        print("End of Test")

    def test_load_tensor(self):
        data = loadTensor("./live-tutorials/inputData/example.tensor", col_types=[int, int])
        self.assertIsInstance(data, TensorData)

    def test_load_tensor2(self):
        data = loadTensor("./live-tutorials/inputData/example_graph.tensor", col_types=[int, int, int])
        self.assertIsInstance(data, TensorData)

    def test_load_gz(self):
        data = loadTensor("./live-tutorials/inputData/example_graph.tensor.gz", col_types=[int, int, int])
        self.assertIsInstance(data, TensorData)

    def test_load_csv(self):
        data = loadTensor("./live-tutorials/inputData/example_time_aggregate.csv")
        self.assertIsInstance(data, TensorData)

    def test_load_multi_file(self):
        data = loadTensor("./live-tutorials/inputData/yelp.edgelist")
        self.assertIsInstance(data, TensorData)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   ioutil.py
@Desc    :   Input and output data function.
'''

# here put the import lib

import gzip
import os
import sys
import pandas as pd
import numpy as np
from . import TensorData
import csv
from .basicutil import set_trace


class File():

    def __init__(self, filename, mode, idxtypes):
        self.filename = filename
        self.mode = mode
        self.idxtypes = idxtypes
        self.dtypes = None
        self.sep = None

    def get_sep_of_file(self):
        '''
        return the separator of the line.
        :param infn: input file
        '''
        sep = None
        fp = open(self.filename, self.mode)
        for line in fp:
            line = line.decode(
                'utf-8') if isinstance(line, bytes) else line
            if (line.startswith("%") or line.startswith("#")):
                continue
            line = line.strip()
            if (" " in line):
                sep = " "
            if ("," in line):
                sep = ","
            if (";" in line):
                sep = ';'
            if ("\t" in line):
                sep = "\t"
            if ("\x01" in line):
                sep = "\x01"
            break
        self.sep = sep

    def transfer_type(self, typex):
        if typex == float:
            _typex = 'float'
        elif typex == int:
            _typex = 'int'
        elif typex == str:
            _typex = 'object'
        else:
            _typex = 'object'
        return _typex

    def _open(self, **kwargs):
        pass

    def _read(self, **kwargs):
        pass


class GZipFile(File):
    def _open(self, **kwargs):
        f = gzip.open(self.filename, self.mode)
        pos = 0
        cur_line = f.readline()
        while cur_line.startswith(b"#"):
            pos = f.tell()
            cur_line = f.readline()
        f.seek(pos)
        _f = open(self.filename, self.mode)
        _f.seek(pos)
        fin = pd.read_csv(f, sep=self.sep, **kwargs)
        column_names = fin.columns
        self.dtypes = {}
        if self.idxtypes is not None:
            for idx, typex in self.idxtypes:
                self.dtypes[column_names[idx]] = self.transfer_type(typex)
            fin = pd.read_csv(_f, dtype=self.dtypes, sep=self.sep, **kwargs)
        else:
            fin = pd.read_csv(_f, sep=self.sep, **kwargs)
        return fin


class TensorFile(File):
    def _open(self, **kwargs):
        if 'r' not in self.mode:
            self.mode += 'r'
        f = open(self.filename, self.mode)
        pos = 0
        cur_line = f.readline()
        while cur_line.startswith('#'):
            pos = f.tell()
            cur_line = f.readline()
        f.seek(pos)
        _f = open(self.filename, self.mode)
        _f.seek(pos)
        fin = pd.read_csv(f, **kwargs)
        column_names = fin.columns
        self.dtypes = {}
        if not self.idxtypes is None:
            for idx, typex in self.idxtypes:
                self.dtypes[column_names[idx]] = self.transfer_type(typex)
            fin = pd.read_csv(_f, dtype=self.dtypes, sep=self.sep, **kwargs)
        else:
            fin = pd.read_csv(_f, sep=self.sep, **kwargs)
        return fin

    def _read(self, **kwargs):
        tensorlist = []
        # self.get_sep_of_file()
        _file = self._open(**kwargs)
        if not self.idxtypes is None:
            idx = [i[0] for i in self.idxtypes]
            tensorlist = _file[idx]
        else:
            tensorlist = _file
        return tensorlist


class CSVFile(File):
    def _open(self, **kwargs):
        f = pd.read_csv(self.filename, **kwargs)
        column_names = list(f.columns)
        self.dtypes = {}
        if not self.idxtypes is None:
            for idx, typex in self.idxtypes:
                self.dtypes[column_names[idx]] = self.transfer_type(typex)
            f = pd.read_csv(self.filename, dtype=self.dtypes, **kwargs)
        else:
            f = pd.read_csv(self.filename, **kwargs)
        return f

    def _read(self, **kwargs):
        tensorlist = pd.DataFrame()
        _file = self._open(**kwargs)
        if not self.idxtypes is None:
            idx = list(self.dtypes.keys())
            tensorlist = _file[idx]
        else:
            tensorlist = _file
        return tensorlist


class NPFile(File):
    def _open(self, **kwargs):
        f = np.load(self.filename)
        return f

    def _read(self, **kwargs):
        f = self._open(**kwargs)
        if self.filename.endswith(".npy"):
            df = pd.DataFrame(f)
        elif self.filename.endswith(".npz"):
            df = pd.DataFrame()
            for k in f.keys():
                df = pd.concat([df, pd.DataFrame(f[k])])
        else:
            df = None
        if not self.idxtypes is None:
            idx = [i[0] for i in self.idxtypes]
            types = [i[1] for i in self.idxtypes]
            df = df[idx]
            for i, _type in enumerate(types):
                df[i] = df[i].astype(self.transfer_type(_type))
        else:
            df = df
        return df


def _read_data(filename: str, idxtypes: list, **kwargs) -> object:
    """Check format of file and read data from file.

    Default format is .tensor. Now we support read from csv, gz, tensor.

    Parameters
    ----------
    filename : str
        file name
    idxtypes : list
        type of columns

    Returns
    ----------
    Data object read from file

    Raises
    ----------
    Exception
        if file cannot be read, raise a FileNotFoundError.
    """

    _class = None
    _postfix = os.path.splitext(filename)[-1]
    mode = 'r'
    if _postfix == ".csv":
        _class = CSVFile
    elif _postfix == ".tensor":
        _class = TensorFile
    elif _postfix in [".npy", ".npz"]:
        _class = NPFile
    elif _postfix in [".gz", ".bz2", ".zip", ".xz"]:
        _class = CSVFile
    else:
        for _postfix in [".csv", ".tensor", ".npy", ".npz", ".gz", ".bz2", ".zip", ".xz"]:
            if os.path.isfile(filename + _postfix):
                _filename = filename + _postfix
                return _read_data(_filename, idxtypes, **kwargs)
        _class = CSVFile
    _obj = _class(filename, mode, idxtypes)
    _data = _obj._read(**kwargs)
    return _data


def _check_compress_file(path: str, cformat=[".gz", ".bz2", ".zip", ".xz"]):
    valpath = None
    if os.path.isfile(path):
        valpath = path
    else:
        for cf in cformat:
            if os.path.isfile(path+cf):
                valpath = path + cf
                return valpath
    if not valpath is None:
        return valpath
    else:
        raise FileNotFoundError(f"{path} cannot be found.")


def _aggregate(data_list):
    if len(data_list) < 1:
        raise Exception("Empty list of data")
    elif len(data_list) == 1:
        return data_list[0]
    else:
        return pd.concat(data_list, axis=0, ignore_index=True)


def _isgzfile(filename):
    return filename.endswith(".gz")


"gzip file must be read and write in binary/bytes"


def _myopenfile(fnm, mode):
    f = None
    if 'w' in mode:
        if _isgzfile(fnm):
            import gzip
            mode = mode+'b' if 'b' != mode[-1] else mode
            f = gzip.open(fnm, mode)
        else:
            f = open(fnm, mode)
    else:
        if 'r' not in mode and 'a' not in mode:
            mode = 'r' + mode
        if os.path.isfile(fnm):
            if not _isgzfile(fnm):
                f = open(fnm, mode)
            else:
                import gzip
                mode = 'rb'
                f = gzip.open(fnm, mode)
        elif os.path.isfile(fnm+'.gz'):
            'file @fnm does not exists, use fnm.gz instead'
            print(
                '==file {} does not exists, read {}.gz instead'.format(fnm,
                                                                       fnm))
            import gzip
            mode = 'rb'
            f = gzip.open(fnm+'.gz', mode)
        else:
            print('file: {} or its zip file does NOT exist'.format(fnm))
            sys.exit(1)
    return f


def loadTensor(path: str,  col_idx: list = None, col_types: list = None, sep: str = ',', **kwargs):
    '''
    Parameters
    ------
        path:
            file-like object
    '''
    if not "header" in kwargs.keys():
        kwargs["header"] = None
    if path is None:
        raise FileNotFoundError('Path is missing.')
    if not "sep" in kwargs.keys():
        kwargs["sep"] = sep

    import glob
    files = glob.glob(path)
    if col_types is None:
        if col_idx is None:
            idxtypes = None
        else:
            idxtypes = [(x, str) for x in col_idx]
    else:
        if col_idx is None:
            col_idx = [i for i in range(len(col_types))]
        if len(col_idx) == len(col_types):
            idxtypes = [[x, col_types[i]] for i, x in enumerate(col_idx)]
        else:
            raise Exception(f"Error: input same size of col_types and col_idx")

    data_list = []
    for _file in files:
        _file = _check_compress_file(_file)
        data_list.append(_read_data(_file, idxtypes, **kwargs))
    data = _aggregate(data_list)
    return TensorData(data)


def loadTensorStream(filename: str, col_idx: list = None, col_types: list = None):
    if not os.path.exists(filename):
        raise FileNotFoundError(f"File {filename} does not exist!")
    if col_types is None:
        if col_idx is None:
            idxtypes = None
        else:
            idxtypes = [(x, str) for x in col_idx]
    else:
        if col_idx is None:
            col_idx = [i for i in range(len(col_types))]
        if len(col_idx) == len(col_types):
            idxtypes = [(x, col_types[i]) for i, x in enumerate(col_idx)]
        else:
            raise Exception(f"Error: input same size of col_types and col_idx")
    postfix = os.path.splitext(filename)[-1]
    if postfix == ".csv":
        f = open(filename, 'r')
        has_header = csv.Sniffer().has_header(f.read(1024))
        f.seek(0)
        if has_header:
            next(f)
    elif postfix == ".tensor":
        f = open(filename, 'r')
    elif postfix in ['.gz', '.bz2', '.zip', '.xz']:
        f = open(filename, 'r')
        has_header = csv.Sniffer().has_header(f.read(1024))
        f.seek(0)
        if has_header:
            next(f)
    else:
        raise FileNotFoundError(
            f"Error: Can not find file {filename}, please check the file!\n")
    return f, idxtypes


def loadFile2Dict(infn: str, n_keyelems: int = 1, key_elem_type: type = int,
                  value_elem_type: type = float, comments: str = "#", delimiter: str = ","):
    '''DESIGNED for EAGLEMINE: load dict data from the input file
    Parameters:
    -------
    :param infn: str
        Input data file
    :param n_keyelems: int
        The number of elements as the key of the 'dict' data, which should be [1, \max{\#elems} -  1]
        Default is 1.
    :param key_elem_type: type
        The type of the key element(s)
        Default is int.
    :param value_elem_type: type
        The type of the value element(s)
        Default is float.
    :param comments: str
        The comments (start character) of inputs.
        Default is "#".
    :param delimiter: str
        The separator of items in each line of inputs.
        Default is ",".
    '''
    data = dict()
    with open(infn, 'r') as fp:
        for line in fp:
            if line.startswith(comments):
                continue
            toks = line.strip().split(delimiter)
            n_elems = len(toks)
            if n_elems <= 1:
                raise ValueError(
                    "Invalid data in input file, which should contain two elements at least")
            if n_keyelems > 1:
                key = tuple(map(key_elem_type, toks[:n_keyelems]))
                if n_elems - n_keyelems > 1:
                    val = tuple(map(value_elem_type, toks[n_keyelems:]))
                else:
                    val = value_elem_type(toks[-1])
                data[key] = val
            elif n_keyelems == 1:
                if n_elems - n_keyelems > 1:
                    val = tuple(map(value_elem_type, toks[1:]))
                else:
                    val = value_elem_type(toks[1])
                data[key_elem_type(toks[0])] = val
            else:
                raise ValueError(
                    "Invalid parameter 'n_keyelems', which should be [{}, {}]".format(1, n_elems - 1))
        fp.close()
    return data


def loadHistogram(infn: str, comments: str = '#', delimiter: str = ','):
    '''Load histogram data
    Parameters:
    --------
    :param infn: str
        The input histogram file with following format:
         the 1st three lines started with ‘#’ are comments for some basic information, that is, 1st line shows the shape of 2-dimensional histogram;
         the 2nd line gives the corresponding real coordinate of each cell coordinate x, and
         the 3rd line is the corresponding real coordinate of each cell coordinate y, these two lines are used for visualizing the histogram.
         Then the followed lines are non-zero data records of the histogram.
    :param comments: str
            The comments (start character) of inputs.
            Default is "#".
        :param delimiter: str
            The separator of items in each line of inputs.
            Default is ",".
    '''
    shape, ticks_dims = list(), list()
    hist_arr = list()

    with open(infn, 'r') as fp:
        line_1st = fp.readline().strip()
        if not line_1st.startswith(comments):
            raise IOError(
                "Invalid input histogram file, which should start with {}".format(comments))
        shape = list(map(int, line_1st[1:].split(delimiter)))

        for m in range(len(shape)):
            line_m = fp.readline().strip()
            if not line_m.startswith(comments):
                raise IOError(
                    "Invalid input histogram file, which should start with {}".format(comments))
            ticks_m = list(map(float, line_m[1:].split(delimiter)))
            ticks_dims.append(ticks_m)

        for line in fp.readlines():
            if line.startswith(comments):
                continue
            toks = line.strip().split(delimiter)
            hist_arr.append(list(map(int, toks)))
        fp.close()

    return np.array(shape, int), ticks_dims, np.array(hist_arr, int)


def saveDictListData(dictls, outdata, delim=':', mode='w'):
    if _isgzfile(outdata):
        'possible mode is ab'
        mode = mode + 'b' if mode[-1] != 'b' else mode
    # write bytes
    ib = True if 'b' in mode else False

    with _myopenfile(outdata, mode) as fw:
        i = 0
        for k, l in dictls.items():
            if not isinstance(l, (list, np.ndarray)):
                print("This is not a dict of value list.", type(l))
                break
            if len(l) < 1:
                continue
            k = k.decode() if isinstance(k, bytes) else str(k)
            ostr = "{}{}".format(k, delim)
            if len(l) < 1:
                i += 1
                continue
            l = [x.decode() if isinstance(x, bytes) else str(x) for x in l]
            ostr = ostr + " ".join(l) + '\n'
            fw.write(ostr.encode() if ib else ostr)
        fw.close()
        if i > 0:
            print("Warn: total {} empty dict lists are removed".format(str(i)))


def loadDictListData(indata, ktype=str, vtype=str, delim=':', mode='r'):
    if _isgzfile(indata):
        mode = 'rb'
    # bytes
    ib = True if 'b' in mode else False

    dictls = {}
    with _myopenfile(indata, mode) as fr:
        for line in fr:
            line = line.decode() if ib else line
            line = line.strip().split(delim)
            lst = []
            for e in line[1].strip().split(' '):
                if vtype == int:
                    lst.append(vtype(float(e)))
                else:
                    lst.append(vtype(e))
            dictls[ktype(line[0].strip())] = lst
        fr.close()
    return dictls



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  Describe hypercubes (or only Two-dimensional) with discrete Gaussian distribution
#  Author: wenchieh
#
#  Project: eaglemine
#      discrete_gaussian.py
#      Version:  1.0
#      Date: November 17 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <11/17/2017>
# 
__author__ = 'wenchieh'

# third-party lib
import numpy as np
from pomegranate import *
from scipy.stats import multivariate_normal


##################################################################################
#
#  A tutorial for package pomegranate is reference to:
#     [1] pomegranate: probabilistic modelling in python 
#         https://homes.cs.washington.edu/~jmschr/lectures/pomegranate.html
#
##################################################################################

class DiscreteGaussian(object):
    def fit_single(self, pos_left, pos_right, weights):
        left, right = np.asarray(pos_left), np.asarray(pos_right)
        centers = (left + right) / 2.0
        fit_model = MultivariateGaussianDistribution.from_samples(centers, weights=weights)
        mu, cov = np.array(fit_model.parameters[0]), np.array(fit_model.parameters[1])

        _pdfs_ = multivariate_normal.logpdf(centers, mu, cov, allow_singular=True)
        loss_log = np.sum([weights[i] * _pdfs_[i] for i in range(len(centers))])
        return [mu], [cov], [1.0], loss_log

    def fit_mixture(self, pos_left, pos_right, weights, n_components=2):
        left, right = np.asarray(pos_left), np.asarray(pos_right)
        centers = (left + right) / 2.0
        fit_gmm = GeneralMixtureModel.from_samples(MultivariateGaussianDistribution,
                                                   n_components=n_components, X=centers, weights=weights,
                                                   stop_threshold=0.01, n_jobs=2)
        mus, covs = list(), list()
        comp_ws = np.array(fit_gmm.weights)
        comp_ws /= np.sum(comp_ws)
        for i in range(n_components):
            paras = fit_gmm.distributions[i].parameters
            mus.append(np.array(paras[0]))
            covs.append(np.array(paras[1]))

        _pdfs_ = fit_gmm.log_probability(centers)
        loss_log = np.sum([weights[i] * _pdfs_[i] for i in range(len(centers))])
        return mus, covs, comp_ws, loss_log


    def log_loss(self, paras, pos_left, pos_right, weights, is_mix, n_components=2):
        left, right = np.asarray(pos_left), np.asarray(pos_right)
        if is_mix is True:
            mus, covs, comp_ws = paras['mus'], paras['covs'], paras['weights']
            n_mus = len(mus)
            if n_mus != n_components:
                print("Warning: input parameter not consistent for mixture Gaussians.")
                print("\t n_component {}, #mu {}s.".format(n_components, len(mus)))
                if n_mus < n_components:
                    print("Error: n_component < #mu (#cov, #weihght), and exit.")
                    exit(0)

            comp_dists = list()
            for i in range(n_components):
                idist = MultivariateGaussianDistribution(mus[i], covs[i])
                comp_dists.append(idist)

            gmm = GeneralMixtureModel(comp_dists, weights=np.array(comp_ws))

            centers = (left + right) / 2.0
            pdfs = gmm.log_probability(centers)
            # loss_log = np.sum([weights[i] * pdfs[i] for i in range(len(centers))])
        else:
            mu, cov = paras['mus'][0], paras['covs'][0]
            centers = (np.array(pos_left) + np.array(pos_right)) / 2.0
            pdfs = multivariate_normal.logpdf(centers, mu, cov, allow_singular=True)

        loss_log = -1.0 * np.sum([weights[i] * pdfs[i] for i in range(len(centers))])
        return loss_log



import numpy as np
from .._model import DMmodel
import scipy.sparse.linalg as slin
import numpy as np
from scipy.sparse import csc_matrix, coo_matrix, csr_matrix, lil_matrix
from spartan.model.fraudar.greedy import logWeightedAveDegree

class Fraudar(DMmodel):
    def __init__(self, data_mat):
        self.data = data_mat

    def run(self, out_path = "./", file_name = "out"):
        sparse_matrix = self.data.to_scipy()
        sparse_matrix = sparse_matrix.asfptype()
        res = logWeightedAveDegree(sparse_matrix)
        print(res)
        np.savetxt("%s.rows" % (out_path + file_name, ), np.array(list(res[0][0])), fmt='%d')
        np.savetxt("%s.cols" % (out_path + file_name, ), np.array(list(res[0][1])), fmt='%d')
        print("score obtained is ", res[1])
    
    
    def anomaly_detection(self):
        return self.run()

    def save(self, outpath):
        pass



# Copyright 2020 The spartan2 Authors.
#
#

import setuptools
from setuptools.extension import Extension
from Cython.Build import cythonize, build_ext

NAME = 'spartan2'
VERSION = '0.1.1'
DESCRIPTION = 'collection of data mining algorithms on big graphs and time series'
URL = 'https://github.com/BGT-M/spartan2'
AUTHOR = 'Shenghua Liu, Houquan Zhou, Quan Ding, Jiabao Zhang, Xin Zhao, Siwei Zeng,etc'
EMAIL = 'liu.shengh@foxmail.com'

# Get the long description from the README file
with open('README.md', encoding='utf-8') as fp:
    _LONG_DESCRIPTION = fp.read()

extensions = [
    Extension(
        name="spartan.model.DPGS.c_MDL",
        sources=["spartan/model/DPGS/c_MDL.pyx"],
    ),
]

REQUIRED_PACKAGES = [
    'numpy',
    'scipy',
    'networkx',
    'matplotlib',
    'statsmodels',
    'pomegranate',
    'scikit-learn',
    'scikit-image',
    'sparse',
    'ipdb',
]

setuptools.setup(name=NAME,
                 version=VERSION,
                 description=DESCRIPTION,
                 long_description=_LONG_DESCRIPTION,
                 long_description_content_type='text/markdown',
                 author=AUTHOR,
                 author_email=EMAIL,
                 url=URL,
                 install_requires=REQUIRED_PACKAGES,
                 packages=setuptools.find_packages(),
                 ext_modules=cythonize(extensions),
                 classifiers=[
                     "Programming Language :: Python :: 3",
                     "License :: OSI Approved :: MIT License",
                     "Operating System :: OS Independent",
                 ],
                 python_requires='>=3.6',
                 build_ext= build_ext
                 )
